import hashlib
import json
from typing import Iterator, List, Tuple

from cognite.client.data_classes import Asset
from cognite.experimental import CogniteClient

from cognite.airworkflow.util.cdf import data_set_id

from .scheduling import (
    Schedule,
    create_schedule_instance,
    create_schedules,
    retrieve_model_assets,
    retrieve_schedule_assets,
    retrieve_updated_schedules,
    update_schedules,
)

AIR_INFRA = "airInfra"
MODEL_POSTFIX = ":latest"
DATA_SET_NAME = "AIR"
EXCLUDED = ["loadbalancer"]


def retrieve_infra_model_assets(client) -> Iterator[Asset]:
    assets = retrieve_model_assets(client, [AIR_INFRA])
    assets = filter(lambda x: x.metadata.get("schedule") and x.external_id not in EXCLUDED, assets)
    return assets


def retrieve_schedule_asset_without_asset(client: CogniteClient) -> List:
    air_assets: List[Asset] = client.assets.list(data_set_ids=[data_set_id(client, DATA_SET_NAME)], limit=-1)
    assets: Iterator[Asset] = retrieve_infra_model_assets(client)
    assets_to_create: List[Asset] = []
    for asset in assets:
        if not any([asset.external_id == x.parent_external_id and "schedule" in x.name for x in air_assets]):
            assets_to_create.append(asset)
    return assets_to_create


def create_asset(client, asset: Asset) -> str:
    data = json.dumps({})
    hashed = hashlib.md5(f"{asset.external_id}-{data}".encode()).hexdigest()  # nosec
    asset_to_create = Asset(
        data_set_id=data_set_id(client, DATA_SET_NAME),
        name=f"{asset.name} schedule",
        external_id=hashed,
        parent_external_id=asset.external_id,
        metadata={"data": data},
    )
    if client.assets.retrieve(external_id=hashed) is None:
        client.assets.create(asset_to_create)
    return hashed


def retrieve_undeployed_infra_schedule_assets(
    client: CogniteClient,
) -> List[Schedule]:
    schedule_assets = retrieve_schedule_assets(client)
    schedules = client.functions.schedules.list()
    schedules_external_ids = map(lambda s: s.description, schedules)
    undeployed_schedule_assets = [i for i in schedule_assets if i.external_id not in schedules_external_ids]
    infra_schedule_assets = list(retrieve_infra_model_assets(client))
    undeployed_infra = filter(
        lambda u: u.parent_external_id in map(lambda a: a.name, infra_schedule_assets),
        undeployed_schedule_assets,
    )
    undeployed_infra_schedules: List[Schedule] = []
    for i in undeployed_infra:
        undeployed_infra_schedules += create_schedule_instance(client, i.external_id)
    return undeployed_infra_schedules


def retrieve_updated_infra_schedules(client: CogniteClient) -> Tuple:
    infra_assets = list(retrieve_infra_model_assets(client))
    updated, to_be_deleted = retrieve_updated_schedules(client)
    updated_filtered = filter(
        lambda x: x.model in map(lambda i: i.external_id, infra_assets),
        updated,
    )
    to_be_deleted_filtered = filter(
        lambda x: x.model in map(lambda i: i.external_id, infra_assets),
        to_be_deleted,
    )
    return list(updated_filtered), list(to_be_deleted_filtered)


def delete_double_schedules(client: CogniteClient, model_names: List):
    schedules = client.functions.schedules.list()
    for model_name in model_names:
        model_schedules = list(filter(lambda s: model_name in s.function_external_id, schedules))
        unique_schedules = set(map(lambda s: s.name, model_schedules))
        for i in unique_schedules:
            extracted = list(filter(lambda s: s.name == i, model_schedules))
            if len(extracted) == 1:
                continue
            max_created = max(map(lambda e: e.created_time, extracted))
            delete_ids = list(filter(lambda e: e.created_time != max_created, extracted))
            [client.functions.schedules.delete(j.id) for j in delete_ids]


def functions_exist_and_ready_external_ids(client: CogniteClient) -> List[str]:
    functions = client.functions.list()
    ready_air_functions = [f.external_id for f in functions if f.owner == "AIR" and f.status == "Ready"]
    return ready_air_functions


def execute(client: CogniteClient):
    missing_schedule_assets = retrieve_schedule_asset_without_asset(client)
    [create_asset(client, asset) for asset in missing_schedule_assets]
    to_be_deployed = retrieve_undeployed_infra_schedule_assets(client)
    existing_air_functions = functions_exist_and_ready_external_ids(client)
    # check if function exists
    existing_deployed = [model for model in to_be_deployed if model.model + MODEL_POSTFIX in existing_air_functions]
    if len(existing_deployed):
        create_schedules(client, existing_deployed)
    updated, to_be_deleted = retrieve_updated_infra_schedules(client)
    existing_updated = [model for model in updated if model.model + MODEL_POSTFIX in existing_air_functions]
    if len(existing_updated):
        update_schedules(client, existing_updated, to_be_deleted)
    print(f"Created: {len(existing_deployed)}, Updated: {len(existing_updated)}")
    infra_assets_names = map(lambda a: a.name, retrieve_infra_model_assets(client))
    infra_function_names = [asset + ":latest" for asset in infra_assets_names]
    delete_double_schedules(client, infra_function_names)



import random
from pathlib import Path

import pandas as pd
from cognite.experimental import CogniteClient

import cognite.airworkflow.constants as const
import cognite.airworkflow.util.cdf as cdf
from cognite.airworkflow.model.config import Config, Schedule
from cognite.airworkflow.model.project import ProjectInfo


class DeploySchedule:
    def __init__(self, project: str, path: Path, config: Config):
        self.info: ProjectInfo = const.AUTHENTICATION[project]
        self.schedule: Schedule = config.schedule
        self.client: CogniteClient = cdf.experimental_client(self.info.project)
        self.model_name: str = path.parts[-1]
        self.model_name_latest: str = self.model_name + ":latest"
        self.functions = self.client.functions.list().to_pandas()

    def has_schedule(self) -> bool:
        schedules = self.client.functions.schedules.list().to_pandas()
        if schedules.shape[0]:
            schedules = schedules.query("functionExternalId.str.contains(@self.model_name_latest)")
            return schedules.shape[0] > 0
        return False

    def retrieve_model_external_id(self, series: pd.Series) -> str:
        return series[series.map(lambda x: x.split("/")[-1] == self.model_name + ":latest")].iloc[0]

    def create_schedule(self, model_external_id: str):
        ran = int(random.random() * 1e5)  # nosec
        return self.client.functions.schedules.create(
            function_external_id=model_external_id,
            name=self.model_name + "-" + str(ran),
            data={},
            description="",
            cron_expression=self.schedule.cron_expression,
        ).name

    def run(self) -> str:
        if not self.schedule.cron_expression:
            return ""

        if self.functions.shape[0] == 0:
            return ""

        if self.functions.query("externalId.str.contains(@self.model_name_latest)").shape[0] == 0:
            return ""

        if self.has_schedule():
            return ""

        model_external_id = self.retrieve_model_external_id(self.functions["externalId"])
        return self.create_schedule(model_external_id)



import json
from typing import Dict, List, Union

PROJECT_FUNCTION_MAP = "ProjectFunctionMap"
PROJECT_PROPERTIES = "Projectproperties"


class Config:
    def __init__(self, config: Dict, function_name: str, repoconfig: Dict):
        self.function_name = function_name
        self.repoconfig = repoconfig

        model_settings = config["modelSettings"]
        model_description = config["modelDescription"]

        self.schedule: Schedule = Schedule(config.get("schedule", {}))

        self.front_end_name: str = model_description["frontEndName"]
        self.description: str = model_description["description"]
        self.long_description: str = model_description["longDescription"]

        self.deploy: List[str] = self._retrieve_deployment_information()
        self.model_version: str = model_settings["modelVersion"]
        self.send_alerts: bool = bool(model_settings["sendAlerts"])
        self.display_in_front_end: bool = bool(model_settings["displayInFrontEnd"])
        self.backfill: bool = bool(model_settings["backfill"])
        self.dependencies: List = model_settings.get("dependencies", [])
        self.air_infrastructure: bool = bool(config.get("airInfra"))
        self.load_balancer: bool = bool(config.get("loadBalancer"))
        self.secret_names: List[str] = model_settings.get("secrets")
        self.fields: Union[Fields, None] = Fields(config.get("fields", {})) if config.get("fields") else None
        self.visualization: Union[Visualization, None] = (
            Visualization(config.get("visualization", {})) if config.get("visualization") else None
        )

    def _retrieve_deployment_information(self) -> List[str]:
        project_function_map = self.repoconfig.get(PROJECT_FUNCTION_MAP, {}).get(self.function_name, [])
        projects = list(self.repoconfig[PROJECT_PROPERTIES].keys())
        return project_function_map or projects

    def to_dict(self):
        convert = {
            "description": self.description,
            "modelVersion": self.model_version,
            "frontEndName": self.front_end_name,
            "sendAlerts": self.send_alerts,
            "longDescription": self.long_description,
            "displayInFrontEnd": self.display_in_front_end,
            "backfill": self.backfill,
        }
        if self.schedule.to_dict():
            convert.update({"schedule": self.schedule.to_dict()})
        if self.dependencies:
            convert.update({"dependencies": self.dependencies})
        if self.air_infrastructure:
            convert.update({"airInfra": self.air_infrastructure})
        if self.load_balancer:
            convert.update({"load_balancer": self.load_balancer})
        if self.fields:
            convert.update({"fields": self.fields.to_dict()})
        if self.visualization:
            convert.update({"visualization": self.visualization.to_dict()})
        return convert

    def __repr__(self):
        return json.dumps(self.to_dict())


class Schedule:
    def __init__(self, schedule: dict):
        self.cron_expression: Union[str, None] = schedule.get("cronExpression")

    def to_dict(self):
        convert = {}
        if self.cron_expression:
            convert.update({"cronExpression": self.cron_expression})
        return convert


class Field:
    def __init__(
        self,
        id: str,
        name: Union[str, None] = None,
        description: Union[str, None] = None,
        type: Union[str, None] = None,
        multiple: Union[bool, None] = None,
        units: Union[List[str], None] = None,
    ):
        self.id: str = id
        self.name: Union[str, None] = name
        self.description: Union[str, None] = description
        self.type: Union[str, None] = type
        self.multiple: Union[bool, None] = multiple
        self.units: Union[List[str], None] = units

    def to_dict(self):
        convert = {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "type": self.type,
        }
        if self.multiple is not None:
            convert.update({"multiple": self.multiple})

        if self.units:
            convert.update({"units": self.units})

        return convert

    def __repr__(self):
        return f"{self.id}, {self.name}, {self.description}, {self.multiple}"


class Fields:
    def __init__(self, fields: Dict):
        self.fields = fields
        self.field_ids: List = list(self.fields.keys())
        self._get_fields: List = []

    @property
    def get_fields(self):
        if self._get_fields:
            return self._get_fields

        local_fields = []
        for field_id in self.field_ids:
            field1 = Field(id=field_id, **self.fields[field_id])
            local_fields.append(field1)
        self._get_fields = local_fields
        return self._get_fields

    def to_dict(self):
        return [field.to_dict() for field in self.get_fields]

    def __repr__(self):
        return " - ".join(self.field_ids)


class Visualization:
    def __init__(self, visualization: Dict):
        self.visualization = visualization
        self.time_series = self.visualization.get("timeSeries", {})
        self.time_series_field_ids = self.time_series.get("fields", [])
        self.thresholds = self.visualization.get("thresholds", {})
        self.thresholds_field_ids = self.thresholds.get("fields", [])
        self.grouping = self.visualization.get("grouping")

    def to_dict(self):
        convert = {
            "timeSeries": {"fields": self.time_series_field_ids},
            "thresholds": {"fields": self.thresholds_field_ids},
        }
        if self.grouping:
            convert["grouping"] = self.grouping
        return convert

    def __repr__(self):
        return self.to_dict()
